// BasIStst.c  

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#define XX 50
//++pac
#define ITER 3000000
// #define M 0
// #define N 32

#define RE 0.005

//++pac
void
version(void)
{
	fprintf(stderr, "  BasIStst\n");
	fprintf(stderr, "  Bayesian Importance Sampler for gene order, test of algorithm\n");
	fprintf(stderr, "  version 0.03\n");
	fprintf(stderr, "  Copy(L)eft 2003 Istvan Miklos\n");
	fprintf(stderr, "  cosmetic changes by Peter A. Cejchan\n");
	exit(-1);
}

//++pac
int main(int argc, char *argv[])
{
  FILE *out;
  int iPermutation[2 * XX + 2]/*, iInvpermutation[2 * XX + 2],iTempperm[2 * XX + 2]*/;
  double eStatistics[2 * XX][2 * XX];
  double eNewtrajectory();
  double eTemp;
  double eSumofweights1, eSumofweightssquared1;
  double eSumofweights2, eSumofweightssquared2;
  double eSumofweightInversion[2 * XX], eSumofweightTransposition[2 * XX];
  int iNumberofproposalsInversion[2 * XX], iNumberofproposalsTransposition[2 * XX];
  double eFirstweight;
  int iIntrand();
  void Swap();
  void Reversion(), Transposition(), Invtransposition1(), Invtransposition2();
  int iI, iJ, iK, iN, iM;
  long int iIteration;
  int M,N;


	switch(argc){
		case 3:
			M=atoi(argv[1]);
			N=atoi(argv[2]);

			printf("m=%d n=%d\n", M, N); fflush(stdout);
			break;
		default:
			version();
	}


  out=fopen("out.txt","w");
//++pac
 #ifdef _PLAN9
  srand(154);
#else
  srandom(154);
#endif



  /* INITIAL PERMUTATION */

  for(iI = 0; iI < 2 * XX + 2; iI++){
    iPermutation[iI]=iI;
  }
  iM = M;
  iN = N;
  while(0 < iN + iM){
    if(iIntrand(iN + iM) < iM){
      iM--;
      iI = iIntrand(XX + 1);
      iJ = iIntrand(XX);
      if(iJ >= iI){
	iJ++;
      }
      else{
	Swap(&iI,&iJ);
      }
      printf("i: %3d, j: %3d\n",iI,iJ);
      Reversion(iPermutation,iI,iJ);
    }
    else{
      iN--;
      iI = iIntrand(XX + 1);
      iJ = iIntrand(XX);
      if(iJ>=iI){
	  iJ++;
      }
      else{
	Swap(&iI,&iJ);
      }
      iK = iIntrand(XX-1);
      if(iK >= iI){
	  iK++;
      }
      if(iK >= iJ){
	  iK++;
      }
      if(iK < iI){
	Swap(&iI,&iK);
	Swap(&iK,&iJ);
      }
      if(iK < iJ){
	Swap(&iJ,&iK);
      }
      printf("i: %3d, j: %3d, k: %3d\n", iI, iJ, iK);
      if(iIntrand(3) == 0){
	Transposition(iPermutation,iI,iJ,iK);
      }
      else{
	if(iIntrand(2) == 0)
	  Invtransposition1(iPermutation,iI,iJ,iK);
	else
	  Invtransposition2(iPermutation,iI,iJ,iK);
      }
    }
  }
  fprintf(out,"***************\n");
  for(iI = 0; iI < 2 * XX + 2; iI++){
    printf("%3d ",iPermutation[iI]);
    fprintf(out,"%3d ",iPermutation[iI]);
  }
  printf("\n");
  fprintf(out,"\n\n");
//++pac
#ifdef _PLAN9
  srand(1025);
#else
  srandom(1025);
#endif

  /* Bayesian importance sampler part */

  /* set the statistics to 0 */

  for(iI = 0; iI < 2 * XX; iI++){
    for(iJ = 0; iJ < 2 * XX; iJ ++){
      eStatistics[iI][iJ] = 0.0;
    }
  }

  for(iI = 0; iI < 2 * XX; iI++){
    eSumofweightInversion[iI] = 0.0; 
    eSumofweightTransposition[iI] = 0.0;
    iNumberofproposalsInversion[iI] = 0;
    iNumberofproposalsTransposition[iI] = 0;
  }


  /* sampling */
  eSumofweights1 = 0.0; 
  eSumofweightssquared1 = 0.0; 
  eSumofweights2 = 0.0; 
  eSumofweightssquared2 = 0.0; 
  iIteration = 0;
  eFirstweight = 0.0;
  while(eFirstweight == 0.0){
    eTemp = 1.0/eNewtrajectory(iPermutation,&iN,&iM);
    if(iN >= 2 * XX || iM >= 2 * XX){
      printf("Too bad rearrangement!\n");
    }
    else{
      for(iI = 1; iI <= iM; iI++){
	eTemp = eTemp * 2.0 / (double)(XX * XX + XX + 2.0);
      }
      eTemp = eTemp * 2.0 / (double)(XX * XX + XX + 2.0);
      for(iJ = 1; iJ <= iN; iJ++){
	eTemp = eTemp * (double)iJ/(double)(iJ + iM);
	eTemp = eTemp * 2.0 / (double) (XX * XX * XX - XX + 2.0);
      }
      eTemp = eTemp * 2.0 / (double) (XX * XX * XX - XX + 2.0);
      eStatistics[iN][iM] += eTemp;
      eFirstweight = eTemp;
      eTemp = 1.0;

      eSumofweightInversion[iM] += eTemp; 
      eSumofweightTransposition[iN] += eTemp;
      iNumberofproposalsInversion[iM]++;
      iNumberofproposalsTransposition[iN]++;

      eSumofweights1 += eTemp *(double) iM ;
      eSumofweightssquared1 += eTemp * eTemp * (double) (iM * iM);
      eSumofweights2 += eTemp *(double) iN ;
      eSumofweightssquared2 += eTemp * eTemp * (double) (iN * iN);
      printf("m = %d n = %d, eSumofweightssquared1 : %e, eSumofweightssquared2 : %e\n",iM,iN,eSumofweightssquared1,eSumofweightssquared2 );
    }
  }
//++pac    while(eSumofweights1 * eSumofweights1 / eSumofweightssquared1 < 1.0/(RE * RE) || eSumofweights2 * eSumofweights2 / eSumofweightssquared2 < 1.0 / (RE * RE)){
while(iIteration < ITER){


    //printf("Iteration %ld.\n",iIteration);
    eTemp = 1.0/(eNewtrajectory(iPermutation,&iN,&iM) * eFirstweight);
    if(iN >= 2 * XX || iM >= 2 * XX){
      printf("Too bad rearrangement!\n");
    }
    else{
      for(iI = 1; iI <= iM; iI++){
	eTemp = eTemp * 2.0 / (double)(XX * XX + XX + 2.0);
      }
      eTemp = eTemp * 2.0 / (double)(XX * XX + XX + 2.0);
      for(iJ = 1; iJ <= iN; iJ++){
	eTemp = eTemp * (double)iJ/(double)(iJ + iM);
	eTemp = eTemp * 2.0 / (double) (XX * XX * XX - XX + 2.0);
      }
      eTemp = eTemp * 2.0 / (double) (XX * XX * XX - XX + 2.0);
      eStatistics[iN][iM] += eTemp;
      if(iNumberofproposalsInversion[iM] > 0){
	eSumofweights1 -= (double) iM * eSumofweightInversion[iM]; 
	eSumofweightssquared1 -= (double) iM * iM *  eSumofweightInversion[iM] *  eSumofweightInversion[iM] / (double) iNumberofproposalsInversion[iM];
      }
      if(iNumberofproposalsTransposition[iN] > 0){
	eSumofweights2 -= (double) iN * eSumofweightTransposition[iN]; 
	eSumofweightssquared2 -= (double) iN * iN *  eSumofweightTransposition[iN] *  eSumofweightTransposition[iN] / (double) iNumberofproposalsTransposition[iN];
      }
      eSumofweightInversion[iM] += eTemp; 
      eSumofweightTransposition[iN] += eTemp;
      iNumberofproposalsInversion[iM]++;
      iNumberofproposalsTransposition[iN]++;

      eSumofweights1 += (double) iM * eSumofweightInversion[iM]; 
      eSumofweightssquared1 += (double) iM * iM *  eSumofweightInversion[iM] *  eSumofweightInversion[iM] / (double) iNumberofproposalsInversion[iM];
      
      eSumofweights2 += (double) iN * eSumofweightTransposition[iN]; 
      eSumofweightssquared2 += (double) iN * iN *  eSumofweightTransposition[iN] *  eSumofweightTransposition[iN] / (double) iNumberofproposalsTransposition[iN];
 
      iIteration++;
      //fprintf(out,"m: %d n: %d, importanceweight: %le ESS1 = %le ESS2 = %le\n",iM,iN,eTemp,eSumofweights1 * eSumofweights1 / eSumofweightssquared1,eSumofweights2 * eSumofweights2 / eSumofweightssquared2);
      printf("ESS1 = %e ESS2 = %e\n",eSumofweights1 * eSumofweights1 / eSumofweightssquared1,eSumofweights2 * eSumofweights2 / eSumofweightssquared2 ); 
      //printf("m = %d n = %d, eSumofweightssquared1 : %le, eSumofweightssquared2 : %le\n",iM,iN,eSumofweightssquared1,eSumofweightssquared2 );

    }
  }
  fclose(out);
  
  eTemp = 0.0;
  for(iI = 0; iI < 2 * XX; iI++){
    for(iJ = 0; iJ < 2 * XX; iJ ++){
      eTemp += eStatistics[iI][iJ];
    }
  }

  for(iI = 0; iI < 2 * XX; iI++){
    for(iJ = 0; iJ < 2 * XX; iJ ++){
      eStatistics[iI][iJ] /= eTemp;
    }
  }

  out = fopen("out","w");
  fprintf(out,"length=  %d, m=  %d, n=  %d, iterations=  %d\n",XX, M, N, ITER);
  for(iI = 0; iI < 2 * XX; iI++){
    for(iJ = 0; iJ < 2 * XX; iJ ++){
      fprintf(out,"%e ",eStatistics[iI][iJ]);
    }
    fprintf(out,"\n");
  }

  return(0);
}

double eNewtrajectory(int iPermutation[2 * XX + 2],
		      int *iN,
		      int *iM){

  double eZ, eNewstep();
  int iTemppermutation[2*XX+2], iI, iSorted;
  int iIntrand();

  *iN=0;
  *iM=0;
  for(iI = 0; iI < 2 * XX + 2; iI++){
    iTemppermutation[iI] = iPermutation[iI];
  }
  iSorted = 0;
  eZ = 1.0;
  while(iSorted == 0){
    eZ = eZ * eNewstep(iTemppermutation,iN,iM);
    //printf("Z: %le\n",eZ);
    iSorted = 1;
    for(iI = 0; iI < 2 * XX + 2; iI++){
      if(iTemppermutation[iI] != iI){
	iSorted=0;
      }
    }
    if(iIntrand(100) == 0 && iSorted == 1){
      eZ *= 0.01;
      iSorted=0;
    }
    
    /*      printf("n: %d, m: %d\n",*n,*m);*/
  }
  eZ = eZ * 0.99;
  // printf("n: %d, m: %d\n",*iN,*iM);
  //scanf("%d\n",&iI);
  return(eZ);
}

double eNewstep(int iPermutation[2 * XX + 2],
		int *iN,
		int *iM){

  int iIntrand();
  void Reversion(), Transposition(), Invtransposition1(), Invtransposition2();
  void Swap();
  int iInvpermutation[2*XX+2];
  int iI, iIi, iJ, iK, iTemp;
  int iNumberofcycles, iCyclelength[XX+2], iNumberofplus[XX+2],/* iInvcycle[XX+2],*/ iSign[XX+2], iTraversed[2*XX+2];
  int p[XX+1], iActualsign[XX+1];
  int iSumsquared, d[XX+1],  iCyclenumber[XX+2];
  int iReverseplus, iReverseminus, iReversesame;
  int iGoodtrans, iGoodinvtrans1, iGoodinvtrans2;
  int iBadtrans, iBadinvtrans1, iBadinvtrans2;
  int iGoodtranslist[5000][3], iGoodinvtranslist1[5000][3], iGoodinvtranslist2[5000][3];
  int iSuperGoodtrans, iSuperGoodinvtrans1, iSuperGoodinvtrans2;
  int iSuperGoodtranslist[5000][3], iSuperGoodinvtranslist1[5000][3], iSuperGoodinvtranslist2[5000][3];
  int iAccepted;
  double eProposal;
  int iGoodall, iSuperGoodall;


  int iKum0[XX + 2], iKumPlus1[XX + 2], iKumMinus11[XX + 2], iKumMinus12[XX + 2];
  int iListofpositives[XX + 2][XX + 2], iListofnegatives[XX + 2][XX + 2];

  /* INVERSE PERMUTATION */

  for(iI = 0; iI < 2 * XX + 2; iI++){
    iInvpermutation[iPermutation[iI]] = iI;
    iTraversed[iI]=0;
    //printf("%3d ",iPermutation[iI]);
  }
  //printf("\n");

  /* DECOMPOSITION TO CYCLES */

  iNumberofcycles=0;
  iIi=0;
  iGoodtrans=0;
  iGoodinvtrans1=0;
  iGoodinvtrans2=0;
  iSuperGoodtrans=0;
  iSuperGoodinvtrans1=0;
  iSuperGoodinvtrans2=0;
  iKum0[0] = 0;
  iKumPlus1[0] = 0;
  iKumMinus11[0] = 0;
  iKumMinus12[0] = 0;

  while(iIi < 2 * XX + 2){
    
    /* finding the begining of the cycle */

    //printf("iIi: %d\n",iIi);    

    while(iTraversed[iIi] == 1 && iIi < 2 * XX + 2){
      iIi++;
    }
   
    //printf("After the while cycle: iIi: %d\n",iIi);    
    
    /* traversing the new cycle*/
    
    if(iIi < 2 * XX + 2){
      iNumberofcycles++;
      iCyclelength[iNumberofcycles - 1] = 0;
      iNumberofplus[iNumberofcycles - 1] = 0;
      iJ = iIi;
    }
    while(iTraversed[iJ] == 0){
      iTraversed[iJ] = 1;
      iCyclelength[iNumberofcycles - 1]++;
      
      /* finding the next black edge */
      
      if(iJ / 2 == (iJ + 1) / 2){
	//	iInvcycle[iJ / 2] = iNumberofcycles;
	iSign[iJ / 2] = 1;
	iCyclenumber[iJ / 2] = iNumberofcycles;
	iActualsign[iCyclelength[iNumberofcycles - 1] - 1] = 1;
	p[iCyclelength[iNumberofcycles - 1] - 1] = iJ / 2;
	iNumberofplus[iNumberofcycles - 1]++;
	iListofpositives[iNumberofcycles -1][iNumberofplus[iNumberofcycles - 1] - 1] = iJ / 2;
	iJ++;
	iTraversed[iJ] = 1;
      }
      else{
	iJ--;
	//	iInvcycle[iJ / 2] = iNumberofcycles;
	iSign[iJ / 2] = -1;
	iCyclenumber[iJ / 2] = iNumberofcycles;
	iActualsign[iCyclelength[iNumberofcycles - 1] - 1] = -1;
	p[iCyclelength[iNumberofcycles - 1] - 1] = iJ / 2;
	iListofnegatives[iNumberofcycles -1][iCyclelength[iNumberofcycles - 1] - iNumberofplus[iNumberofcycles - 1]-1] = iJ / 2;
	iTraversed[iJ] = 1;
      }
      if(iPermutation[iJ] / 2 == (iPermutation[iJ] + 1) / 2 ){
	iJ = iInvpermutation[iPermutation[iJ] + 1];
      }
      else{
	iJ = iInvpermutation[iPermutation[iJ] - 1];
      }
      
    }

    /* Obtaining the kumulative functions */ 
    
    iKum0[iNumberofcycles - 1] = iNumberofplus[iNumberofcycles - 1]*(iNumberofplus[iNumberofcycles - 1] - 1) + (iCyclelength[iNumberofcycles - 1] - iNumberofplus[iNumberofcycles - 1])*(iCyclelength[iNumberofcycles - 1] - iNumberofplus[iNumberofcycles - 1] - 1);
    iKumPlus1[iNumberofcycles - 1] = iNumberofplus[iNumberofcycles - 1]*(iCyclelength[iNumberofcycles - 1] -iNumberofplus[iNumberofcycles - 1]);
    iKumMinus11[iNumberofcycles - 1] = iCyclelength[iNumberofcycles - 1]*(XX + 1 - iCyclelength[iNumberofcycles - 1]);
    iKumMinus12[iNumberofcycles - 1] = iCyclelength[iNumberofcycles - 1];
    if(iNumberofcycles > 1){
      iKum0[iNumberofcycles - 1] += iKum0[iNumberofcycles - 2];
      iKumPlus1[iNumberofcycles - 1] += iKumPlus1[iNumberofcycles - 2];
      iKumMinus11[iNumberofcycles - 1] += iKumMinus11[iNumberofcycles - 2];
      iKumMinus12[iNumberofcycles - 1] += iKumMinus12[iNumberofcycles - 2];
    }
    
    /* investigating long cycles */
    
    if(iCyclelength[iNumberofcycles - 1] > 2){
      //printf("Long cycle!, cyclelength: %d\n",iCyclelength[iNumberofcycles - 1]);
      for(iI = 0; iI < iCyclelength[iNumberofcycles - 1] - 2; iI++){
	for(iJ = iI + 1; iJ < iCyclelength[iNumberofcycles - 1] - 1; iJ++){
	  for(iK = iJ + 1; iK < iCyclelength[iNumberofcycles - 1]; iK++){
	    if(p[iI] < p[iJ] && p[iJ] < p[iK]){
	      if(iActualsign[iI] + iActualsign[iJ] + iActualsign[iK] == -3){
		iSuperGoodtranslist[iSuperGoodtrans][0] = p[iI];
		iSuperGoodtranslist[iSuperGoodtrans][1] = p[iJ];
		iSuperGoodtranslist[iSuperGoodtrans][2] = p[iK];
		iSuperGoodtrans++;
		iGoodinvtranslist1[iGoodinvtrans1][0] = p[iI];
		iGoodinvtranslist1[iGoodinvtrans1][1] = p[iJ];
		iGoodinvtranslist1[iGoodinvtrans1][2] = p[iK];
		iGoodinvtrans1++;
		iGoodinvtranslist2[iGoodinvtrans2][0] = p[iI];
		iGoodinvtranslist2[iGoodinvtrans2][1] = p[iJ];
		iGoodinvtranslist2[iGoodinvtrans2][2] = p[iK];
		iGoodinvtrans2++;
	      }
	      
	      if(iActualsign[iI]==1&&iActualsign[iJ]==-1&&iActualsign[iK]==-1){
		iGoodtranslist[iGoodtrans][0] = p[iI];
		iGoodtranslist[iGoodtrans][1] = p[iJ];
		iGoodtranslist[iGoodtrans][2] = p[iK];
		iGoodtrans++;
	      }
	      
	      if(iActualsign[iI]==-1&&iActualsign[iJ]*iActualsign[iK]==-1){
		iGoodtranslist[iGoodtrans][0] = p[iI];
		iGoodtranslist[iGoodtrans][1] = p[iJ];
		iGoodtranslist[iGoodtrans][2] = p[iK];
		iGoodtrans++;
	      }
	      
	      if((iActualsign[iI]==1) && (iActualsign[iJ]==1) && (iActualsign[iK]==-1)){
		iGoodinvtranslist2[iGoodinvtrans2][0] = p[iI];
		iGoodinvtranslist2[iGoodinvtrans2][1] = p[iJ];
		iGoodinvtranslist2[iGoodinvtrans2][2] = p[iK];
		iGoodinvtrans2++;
	      }
	      if((iActualsign[iI]==1) && (iActualsign[iJ]==-1) && (iActualsign[iK]==-1)){
		iSuperGoodinvtranslist2[iSuperGoodinvtrans2][0] = p[iI];
		iSuperGoodinvtranslist2[iSuperGoodinvtrans2][1] = p[iJ];
		iSuperGoodinvtranslist2[iSuperGoodinvtrans2][2] = p[iK];
		iSuperGoodinvtrans2++;
	      }
	      
	      
	      if((iActualsign[iI]==-1) && (iActualsign[iJ]==1) && (iActualsign[iK]==1)){
		iGoodinvtranslist1[iGoodinvtrans1][0] = p[iI];
		iGoodinvtranslist1[iGoodinvtrans1][1] = p[iJ];
		iGoodinvtranslist1[iGoodinvtrans1][2] = p[iK];
		iGoodinvtrans1++;
	      }

	      if((iActualsign[iI]==-1) && (iActualsign[iJ]==-1) && (iActualsign[iK]==1)){
		iSuperGoodinvtranslist1[iSuperGoodinvtrans1][0] = p[iI];
		iSuperGoodinvtranslist1[iSuperGoodinvtrans1][1] = p[iJ];
		iSuperGoodinvtranslist1[iSuperGoodinvtrans1][2] = p[iK];
		iSuperGoodinvtrans1++;
	      }
	      
	      if((iActualsign[iI]==1) && (iActualsign[iJ]==-1) && (iActualsign[iK]==1)){
		iGoodinvtranslist1[iGoodinvtrans1][0] = p[iI];
		iGoodinvtranslist1[iGoodinvtrans1][1] = p[iJ];
		iGoodinvtranslist1[iGoodinvtrans1][2] = p[iK];
		iGoodinvtrans1++;
		iGoodinvtranslist2[iGoodinvtrans2][0] = p[iI];
		iGoodinvtranslist2[iGoodinvtrans2][1] = p[iJ];
		iGoodinvtranslist2[iGoodinvtrans2][2] = p[iK];
		iGoodinvtrans2++;
	      }
	      
	    }
	    
	    if(p[iI]<p[iK]&&p[iK]<p[iJ]){
	      if(iActualsign[iI]+iActualsign[iJ]+iActualsign[iK]==3){
		iSuperGoodtranslist[iSuperGoodtrans][0] = p[iI];
		iSuperGoodtranslist[iSuperGoodtrans][1] = p[iJ];
		iSuperGoodtranslist[iSuperGoodtrans][2] = p[iK];
		iSuperGoodtrans++;
		iGoodinvtranslist1[iGoodinvtrans1][0] = p[iI];
		iGoodinvtranslist1[iGoodinvtrans1][1] = p[iJ];
		iGoodinvtranslist1[iGoodinvtrans1][2] = p[iK];
		iGoodinvtrans1++;
		iGoodinvtranslist2[iGoodinvtrans2][0] = p[iI];
		iGoodinvtranslist2[iGoodinvtrans2][1] = p[iJ];
		iGoodinvtranslist2[iGoodinvtrans2][2] = p[iK];
		iGoodinvtrans2++;
	      }
	      
	      if(iActualsign[iI]==-1&&iActualsign[iJ]==1&&iActualsign[iK]==1){
		iGoodtranslist[iGoodtrans][0] = p[iI];
		iGoodtranslist[iGoodtrans][1] = p[iJ];
		iGoodtranslist[iGoodtrans][2] = p[iK];
		iGoodtrans++;
	      }
	      
	      if(iActualsign[iI]==1&&iActualsign[iJ]*iActualsign[iK]==-1){
		iGoodtranslist[iGoodtrans][0] = p[iI];
		iGoodtranslist[iGoodtrans][1] = p[iJ];
		iGoodtranslist[iGoodtrans][2] = p[iK];
		iGoodtrans++;
	      }		    		      
	      
	      if(iActualsign[iI]==-1&&iActualsign[iJ]==1&&iActualsign[iK]==-1){
		iGoodinvtranslist2[iGoodinvtrans2][0] = p[iI];
		iGoodinvtranslist2[iGoodinvtrans2][1] = p[iJ];
		iGoodinvtranslist2[iGoodinvtrans2][2] = p[iK];
		iGoodinvtrans2++;
	      }
	      if(iActualsign[iI]==-1&&iActualsign[iJ]==1&&iActualsign[iK]==1){
		iSuperGoodinvtranslist2[iSuperGoodinvtrans2][0] = p[iI];
		iSuperGoodinvtranslist2[iSuperGoodinvtrans2][1] = p[iJ];
		iSuperGoodinvtranslist2[iSuperGoodinvtrans2][2] = p[iK];
		iSuperGoodinvtrans2++;
	      }
	      
	      
	      if(iActualsign[iI]==1&&iActualsign[iJ]==-1&&iActualsign[iK]==-1){
		iGoodinvtranslist1[iGoodinvtrans1][0] = p[iI];
		iGoodinvtranslist1[iGoodinvtrans1][1] = p[iJ];
		iGoodinvtranslist1[iGoodinvtrans1][2] = p[iK];
		iGoodinvtrans1++;
	      }
	      if(iActualsign[iI]==1&&iActualsign[iJ]==-1&&iActualsign[iK]==1){
		iSuperGoodinvtranslist1[iSuperGoodinvtrans1][0] = p[iI];
		iSuperGoodinvtranslist1[iSuperGoodinvtrans1][1] = p[iJ];
		iSuperGoodinvtranslist1[iSuperGoodinvtrans1][2] = p[iK];
		iSuperGoodinvtrans1++;
	      }
	      
	      if((iActualsign[iI]==-1) && (iActualsign[iJ]==-1) && (iActualsign[iK]==1)){
		iGoodinvtranslist1[iGoodinvtrans1][0] = p[iI];
		iGoodinvtranslist1[iGoodinvtrans1][1] = p[iJ];
		iGoodinvtranslist1[iGoodinvtrans1][2] = p[iK];
		iGoodinvtrans1++;
		iGoodinvtranslist2[iGoodinvtrans2][0] = p[iI];
		iGoodinvtranslist2[iGoodinvtrans2][1] = p[iJ];
		iGoodinvtranslist2[iGoodinvtrans2][2] = p[iK];
		iGoodinvtrans2++;
	      }
	    }
	    if(p[iJ]<p[iI]&&p[iI]<p[iK]){
	      if(iActualsign[iI]+iActualsign[iJ]+iActualsign[iK]==3){
		iSuperGoodtranslist[iSuperGoodtrans][0] = p[iI];
		iSuperGoodtranslist[iSuperGoodtrans][1] = p[iJ];
		iSuperGoodtranslist[iSuperGoodtrans][2] = p[iK];
		iSuperGoodtrans++;
		iGoodinvtranslist1[iGoodinvtrans1][0] = p[iI];
		iGoodinvtranslist1[iGoodinvtrans1][1] = p[iJ];
		iGoodinvtranslist1[iGoodinvtrans1][2] = p[iK];
		iGoodinvtrans1++;
		iGoodinvtranslist2[iGoodinvtrans2][0] = p[iI];
		iGoodinvtranslist2[iGoodinvtrans2][1] = p[iJ];
		iGoodinvtranslist2[iGoodinvtrans2][2] = p[iK];
		iGoodinvtrans2++;
	      }
	      
	      if(iActualsign[iI]==1&&iActualsign[iJ]==-1&&iActualsign[iK]==1){
		iGoodtranslist[iGoodtrans][0] = p[iI];
		iGoodtranslist[iGoodtrans][1] = p[iJ];
		iGoodtranslist[iGoodtrans][2] = p[iK];
		iGoodtrans++;
	      }
	      
	      if(iActualsign[iJ]==1&&iActualsign[iI]*iActualsign[iK]==-1){
		iGoodtranslist[iGoodtrans][0] = p[iI];
		iGoodtranslist[iGoodtrans][1] = p[iJ];
		iGoodtranslist[iGoodtrans][2] = p[iK];
		iGoodtrans++;
	      }
	      
	      if(iActualsign[iJ]==-1&&iActualsign[iK]==1&&iActualsign[iI]==-1){
		iGoodinvtranslist2[iGoodinvtrans2][0] = p[iI];
		iGoodinvtranslist2[iGoodinvtrans2][1] = p[iJ];
		iGoodinvtranslist2[iGoodinvtrans2][2] = p[iK];
		iGoodinvtrans2++;
	      }
	      if(iActualsign[iJ]==-1&&iActualsign[iK]==1&&iActualsign[iI]==1){
		iSuperGoodinvtranslist2[iSuperGoodinvtrans2][0] = p[iI];
		iSuperGoodinvtranslist2[iSuperGoodinvtrans2][1] = p[iJ];
		iSuperGoodinvtranslist2[iSuperGoodinvtrans2][2] = p[iK];
		iSuperGoodinvtrans2++;
	      }
	      
	      
	      if(iActualsign[iJ]==1&&iActualsign[iK]==-1&&iActualsign[iI]==-1){
		iGoodinvtranslist1[iGoodinvtrans1][0] = p[iI];
		iGoodinvtranslist1[iGoodinvtrans1][1] = p[iJ];
		iGoodinvtranslist1[iGoodinvtrans1][2] = p[iK];
		iGoodinvtrans1++;
	      }
	      if(iActualsign[iJ]==1&&iActualsign[iK]==-1&&iActualsign[iI]==1){
		iSuperGoodinvtranslist1[iSuperGoodinvtrans1][0] = p[iI];
		iSuperGoodinvtranslist1[iSuperGoodinvtrans1][1] = p[iJ];
		iSuperGoodinvtranslist1[iSuperGoodinvtrans1][2] = p[iK];
		iSuperGoodinvtrans1++;
	      }
	      
	      if((iActualsign[iI]==1) && (iActualsign[iJ]==-1) && (iActualsign[iK]==-1)){
		iGoodinvtranslist1[iGoodinvtrans1][0] = p[iI];
		iGoodinvtranslist1[iGoodinvtrans1][1] = p[iJ];
		iGoodinvtranslist1[iGoodinvtrans1][2] = p[iK];
		iGoodinvtrans1++;
		iGoodinvtranslist2[iGoodinvtrans2][0] = p[iI];
		iGoodinvtranslist2[iGoodinvtrans2][1] = p[iJ];
		iGoodinvtranslist2[iGoodinvtrans2][2] = p[iK];
		iGoodinvtrans2++;
	      }
	      
	    }
	    if(p[iJ]<p[iK]&&p[iK]<p[iI]){
	      if(iActualsign[iI]+iActualsign[iJ]+iActualsign[iK]==-3){
		iSuperGoodtranslist[iSuperGoodtrans][0] = p[iI];
		iSuperGoodtranslist[iSuperGoodtrans][1] = p[iJ];
		iSuperGoodtranslist[iSuperGoodtrans][2] = p[iK];
		iSuperGoodtrans++;
		iGoodinvtranslist1[iGoodinvtrans1][0] = p[iI];
		iGoodinvtranslist1[iGoodinvtrans1][1] = p[iJ];
		iGoodinvtranslist1[iGoodinvtrans1][2] = p[iK];
		iGoodinvtrans1++;
		iGoodinvtranslist2[iGoodinvtrans2][0] = p[iI];
		iGoodinvtranslist2[iGoodinvtrans2][1] = p[iJ];
		iGoodinvtranslist2[iGoodinvtrans2][2] = p[iK];
		iGoodinvtrans2++;
	      }
	      
	      if(iActualsign[iI]==-1&&iActualsign[iJ]==1&&iActualsign[iK]==-1){
		iGoodtranslist[iGoodtrans][0] = p[iI];
		iGoodtranslist[iGoodtrans][1] = p[iJ];
		iGoodtranslist[iGoodtrans][2] = p[iK];
		iGoodtrans++;
	      }
	      
	      if(iActualsign[iJ]==-1&&iActualsign[iI]*iActualsign[iK]==-1){
		iGoodtranslist[iGoodtrans][0] = p[iI];
		iGoodtranslist[iGoodtrans][1] = p[iJ];
		iGoodtranslist[iGoodtrans][2] = p[iK];
		iGoodtrans++;
	      }
	      
	      if(iActualsign[iI]==-1&&iActualsign[iJ]==1&&iActualsign[iK]==1){
		iGoodinvtranslist2[iGoodinvtrans2][0] = p[iI];
		iGoodinvtranslist2[iGoodinvtrans2][1] = p[iJ];
		iGoodinvtranslist2[iGoodinvtrans2][2] = p[iK];
		iGoodinvtrans2++;
	      }
	      if(iActualsign[iI]==-1&&iActualsign[iJ]==1&&iActualsign[iK]==-1){
		iSuperGoodinvtranslist2[iSuperGoodinvtrans2][0] = p[iI];
		iSuperGoodinvtranslist2[iSuperGoodinvtrans2][1] = p[iJ];
		iSuperGoodinvtranslist2[iSuperGoodinvtrans2][2] = p[iK];
		iSuperGoodinvtrans2++;
	      }
	      
	      
	      if(iActualsign[iI]==1&&iActualsign[iJ]==-1&&iActualsign[iK]==1){
		iGoodinvtranslist1[iGoodinvtrans1][0] = p[iI];
		iGoodinvtranslist1[iGoodinvtrans1][1] = p[iJ];
		iGoodinvtranslist1[iGoodinvtrans1][2] = p[iK];
		iGoodinvtrans1++;
	      }
	      if(iActualsign[iI]==1&&iActualsign[iJ]==-1&&iActualsign[iK]==-1){
		iSuperGoodinvtranslist1[iSuperGoodinvtrans1][0] = p[iI];
		iSuperGoodinvtranslist1[iSuperGoodinvtrans1][1] = p[iJ];
		iSuperGoodinvtranslist1[iSuperGoodinvtrans1][2] = p[iK];
		iSuperGoodinvtrans1++;
	      }
	      
	      if((iActualsign[iI]==1) && (iActualsign[iJ]==1) && (iActualsign[iK]==-1)){
		iGoodinvtranslist1[iGoodinvtrans1][0] = p[iI];
		iGoodinvtranslist1[iGoodinvtrans1][1] = p[iJ];
		iGoodinvtranslist1[iGoodinvtrans1][2] = p[iK];
		iGoodinvtrans1++;
		iGoodinvtranslist2[iGoodinvtrans2][0] = p[iI];
		iGoodinvtranslist2[iGoodinvtrans2][1] = p[iJ];
		iGoodinvtranslist2[iGoodinvtrans2][2] = p[iK];
		iGoodinvtrans2++;
	      }
	      
	    }
	    if(p[iK]<p[iI]&&p[iI]<p[iJ]){
	      if(iActualsign[iI]+iActualsign[iJ]+iActualsign[iK]==-3){
		iSuperGoodtranslist[iSuperGoodtrans][0] = p[iI];
		iSuperGoodtranslist[iSuperGoodtrans][1] = p[iJ];
		iSuperGoodtranslist[iSuperGoodtrans][2] = p[iK];
		iSuperGoodtrans++;
		iGoodinvtranslist1[iGoodinvtrans1][0] = p[iI];
		iGoodinvtranslist1[iGoodinvtrans1][1] = p[iJ];
		iGoodinvtranslist1[iGoodinvtrans1][2] = p[iK];
		iGoodinvtrans1++;
		iGoodinvtranslist2[iGoodinvtrans2][0] = p[iI];
		iGoodinvtranslist2[iGoodinvtrans2][1] = p[iJ];
		iGoodinvtranslist2[iGoodinvtrans2][2] = p[iK];
		iGoodinvtrans2++;
	      }			
	      
	      if(iActualsign[iI]==-1&&iActualsign[iJ]==-1&&iActualsign[iK]==1){
		iGoodtranslist[iGoodtrans][0] = p[iI];
		iGoodtranslist[iGoodtrans][1] = p[iJ];
		iGoodtranslist[iGoodtrans][2] = p[iK];
		iGoodtrans++;
	      }
	      
	      if(iActualsign[iK]==-1&&iActualsign[iI]*iActualsign[iJ]==-1){
		iGoodtranslist[iGoodtrans][0] = p[iI];
		iGoodtranslist[iGoodtrans][1] = p[iJ];
		iGoodtranslist[iGoodtrans][2] = p[iK];
		iGoodtrans++;
	      }
	      
	      if(iActualsign[iJ]==-1&&iActualsign[iK]==1&&iActualsign[iI]==1){
		iGoodinvtranslist2[iGoodinvtrans2][0] = p[iI];
		iGoodinvtranslist2[iGoodinvtrans2][1] = p[iJ];
		iGoodinvtranslist2[iGoodinvtrans2][2] = p[iK];
		iGoodinvtrans2++;
	      }
	      if(iActualsign[iJ]==-1&&iActualsign[iK]==1&&iActualsign[iI]==-1){
		iSuperGoodinvtranslist2[iSuperGoodinvtrans2][0] = p[iI];
		iSuperGoodinvtranslist2[iSuperGoodinvtrans2][1] = p[iJ];
		iSuperGoodinvtranslist2[iSuperGoodinvtrans2][2] = p[iK];
		iSuperGoodinvtrans2++;
	      }
	      
	      
	      if(iActualsign[iJ]==1&&iActualsign[iK]==-1&&iActualsign[iI]==1){
		iGoodinvtranslist1[iGoodinvtrans1][0] = p[iI];
		iGoodinvtranslist1[iGoodinvtrans1][1] = p[iJ];
		iGoodinvtranslist1[iGoodinvtrans1][2] = p[iK];
		iGoodinvtrans1++;
	      }
	      if(iActualsign[iJ]==1&&iActualsign[iK]==-1&&iActualsign[iI]==-11){
		iSuperGoodinvtranslist1[iSuperGoodinvtrans1][0] = p[iI];
		iSuperGoodinvtranslist1[iSuperGoodinvtrans1][1] = p[iJ];
		iSuperGoodinvtranslist1[iSuperGoodinvtrans1][2] = p[iK];
		iSuperGoodinvtrans1++;
	      }
	      
	      if((iActualsign[iI]==-1) && (iActualsign[iJ]==1) && (iActualsign[iK]==1)){
		iGoodinvtranslist1[iGoodinvtrans1][0] = p[iI];
		iGoodinvtranslist1[iGoodinvtrans1][1] = p[iJ];
		iGoodinvtranslist1[iGoodinvtrans1][2] = p[iK];
		iGoodinvtrans1++;
		iGoodinvtranslist2[iGoodinvtrans2][0] = p[iI];
		iGoodinvtranslist2[iGoodinvtrans2][1] = p[iJ];
		iGoodinvtranslist2[iGoodinvtrans2][2] = p[iK];
		iGoodinvtrans2++;
	      }
	    }
	    if(p[iK]<p[iJ]&&p[iJ]<p[iI]){
	      if(iActualsign[iI]+iActualsign[iJ]+iActualsign[iK]==3)
		{
		  iSuperGoodtranslist[iSuperGoodtrans][0] = p[iI];
		  iSuperGoodtranslist[iSuperGoodtrans][1] = p[iJ];
		  iSuperGoodtranslist[iSuperGoodtrans][2] = p[iK];
		  iSuperGoodtrans++;
		  iGoodinvtranslist1[iGoodinvtrans1][0] = p[iI];
		  iGoodinvtranslist1[iGoodinvtrans1][1] = p[iJ];
		  iGoodinvtranslist1[iGoodinvtrans1][2] = p[iK];
		  iGoodinvtrans1++;
		  iGoodinvtranslist2[iGoodinvtrans2][0] = p[iI];
		  iGoodinvtranslist2[iGoodinvtrans2][1] = p[iJ];
		  iGoodinvtranslist2[iGoodinvtrans2][2] = p[iK];
		  iGoodinvtrans2++;
		}
	      
	      if(iActualsign[iI]==1&&iActualsign[iJ]==1&&iActualsign[iK]==-1){
		iGoodtranslist[iGoodtrans][0] = p[iI];
		iGoodtranslist[iGoodtrans][1] = p[iJ];
		iGoodtranslist[iGoodtrans][2] = p[iK];
		iGoodtrans++;
	      }
	      
	      if(iActualsign[iK]==1&&iActualsign[iI]*iActualsign[iJ]==-1){
		iGoodtranslist[iGoodtrans][0] = p[iI];
		iGoodtranslist[iGoodtrans][1] = p[iJ];
		iGoodtranslist[iGoodtrans][2] = p[iK];
		iGoodtrans++;
	      }
	      
	      if(iActualsign[iI]==1&&iActualsign[iK]==-1&&iActualsign[iJ]==-1){
		iGoodinvtranslist2[iGoodinvtrans2][0] = p[iI];
		iGoodinvtranslist2[iGoodinvtrans2][1] = p[iJ];
		iGoodinvtranslist2[iGoodinvtrans2][2] = p[iK];
		iGoodinvtrans2++;
	      }
	      if(iActualsign[iI]==1&&iActualsign[iK]==-1&&iActualsign[iJ]==1){
		iSuperGoodinvtranslist2[iSuperGoodinvtrans2][0] = p[iI];
		iSuperGoodinvtranslist2[iSuperGoodinvtrans2][1] = p[iJ];
		iSuperGoodinvtranslist2[iSuperGoodinvtrans2][2] = p[iK];
		iSuperGoodinvtrans2++;
	      }
	      
	      
	      if(iActualsign[iI]==-1&&iActualsign[iK]==1&&iActualsign[iJ]==-1){
		iGoodinvtranslist1[iGoodinvtrans1][0] = p[iI];
		iGoodinvtranslist1[iGoodinvtrans1][1] = p[iJ];
		iGoodinvtranslist1[iGoodinvtrans1][2] = p[iK];
		iGoodinvtrans1++;
	      }
	      if(iActualsign[iI]==-1&&iActualsign[iK]==1&&iActualsign[iJ]==1){
		iSuperGoodinvtranslist1[iSuperGoodinvtrans1][0] = p[iI];
		iSuperGoodinvtranslist1[iSuperGoodinvtrans1][1] = p[iJ];
		iSuperGoodinvtranslist1[iSuperGoodinvtrans1][2] = p[iK];
		iSuperGoodinvtrans1++;
	      }
	      
	      if((iActualsign[iI]==-1) && (iActualsign[iJ]==1) && (iActualsign[iK]==-1)){
		iGoodinvtranslist1[iGoodinvtrans1][0] = p[iI];
		iGoodinvtranslist1[iGoodinvtrans1][1] = p[iJ];
		iGoodinvtranslist1[iGoodinvtrans1][2] = p[iK];
		iGoodinvtrans1++;
		iGoodinvtranslist2[iGoodinvtrans2][0] = p[iI];			  
		iGoodinvtranslist2[iGoodinvtrans2][1] = p[iJ];
		iGoodinvtranslist2[iGoodinvtrans2][2] = p[iK];
		iGoodinvtrans2++;
	      }
	      
	    }
	  }
	}
      }
    }
  }

  
  
  iSumsquared=0;
  iReverseplus=0;
  for(iI = 0; iI < iNumberofcycles; iI++){
    iSumsquared+=iCyclelength[iI]*(iCyclelength[iI]-1);
    d[iI]=(iCyclelength[iI]-iNumberofplus[iI])*iNumberofplus[iI];
    iReverseplus+=d[iI];
  }
  iReverseminus=((XX+1)*XX-iSumsquared)/2;
  iReversesame=(XX+1)*XX/2-iReverseplus-iReverseminus;
  
  /* PROPOSING A NEW STEP */

  eProposal = 1.0;

  iGoodall = iGoodtrans + iGoodinvtrans1 + iGoodinvtrans2;
  iSuperGoodall = iSuperGoodtrans + iSuperGoodinvtrans1 + iSuperGoodinvtrans2;

  /* bening mutation */
  if(iGoodall + iSuperGoodall + iReverseplus > 0 && iIntrand(100) != 0){
    eProposal *= 0.99;
    /* 2-transposition */
    if(iSuperGoodall > 0 && iIntrand(100) < 50){
      eProposal *= 0.5/(double)iSuperGoodall;
      iTemp = iIntrand(iSuperGoodall);
      if(iTemp < iSuperGoodtrans){
	iTemp = iIntrand(iSuperGoodtrans);
	iI = iSuperGoodtranslist[iTemp][0];
	iJ = iSuperGoodtranslist[iTemp][1];
	iK = iSuperGoodtranslist[iTemp][2];
	if(iI > iJ){
	  Swap(&iI,&iJ);
	}
	if(iI > iK){
	  Swap(&iI,&iK);
	}
	if(iJ > iK){
	  Swap(&iJ,&iK);
	}
	Transposition(iPermutation,iI,iJ,iK);
	//printf("Transposition, Super i: %d, j: %d, k: %d\n",iI,iJ,iK);
	  (*iN)++;
      }
      else if(iTemp < iSuperGoodtrans + iSuperGoodinvtrans1){
	iTemp = iIntrand(iSuperGoodinvtrans1);
	iI = iSuperGoodinvtranslist1[iTemp][0];
	iJ = iSuperGoodinvtranslist1[iTemp][1];
	iK = iSuperGoodinvtranslist1[iTemp][2];
	if(iI > iJ){
	  Swap(&iI,&iJ);
	}
	if(iI > iK){
	  Swap(&iI,&iK);
	}
	if(iJ > iK){
	  Swap(&iJ,&iK);
	}
	Invtransposition1(iPermutation,iI,iJ,iK);
	//printf("Invtransposition1, Super i: %d, j: %d, k: %d\n",iI,iJ,iK);
	(*iN)++;
      }
      else{
	iTemp = iIntrand(iSuperGoodinvtrans2);
	iI = iSuperGoodinvtranslist2[iTemp][0];
	iJ = iSuperGoodinvtranslist2[iTemp][1];
	iK = iSuperGoodinvtranslist2[iTemp][2];
	if(iI > iJ){
	  Swap(&iI,&iJ);
	}
	if(iI > iK){
	  Swap(&iI,&iK);
	}
	if(iJ > iK){
	  Swap(&iJ,&iK);
	}
	Invtransposition2(iPermutation,iI,iJ,iK);
	//printf("Invtransposition2, Super i: %d, j: %d, k: %d\n",iI,iJ,iK);
	(*iN)++;
      }
    }
    /* 1-transposition or 1-inversion */
    else{
      if(iSuperGoodall != 0){
	eProposal *= 0.5;
      }
      eProposal /= (double) (iSuperGoodall + iReverseplus);
      iTemp = iIntrand(iSuperGoodall + iReverseplus);
      if(iTemp < iGoodtrans){
	iTemp = iIntrand(iGoodtrans);
	iI = iGoodtranslist[iTemp][0];
	iJ = iGoodtranslist[iTemp][1];
	iK = iGoodtranslist[iTemp][2];
	if(iI > iJ){
	  Swap(&iI,&iJ);
	}
	if(iI > iK){
	  Swap(&iI,&iK);
	}
	if(iJ > iK){
	  Swap(&iJ,&iK);
	}
	//printf("Transposition, good i: %d, j: %d, k: %d\n",iI,iJ,iK);
	Transposition(iPermutation,iI,iJ,iK);
	(*iN)++;
      }
      else if(iTemp < iGoodtrans + iGoodinvtrans1){
	iTemp = iIntrand(iGoodinvtrans1);
	iI = iGoodinvtranslist1[iTemp][0];
	iJ = iGoodinvtranslist1[iTemp][1];
	iK = iGoodinvtranslist1[iTemp][2];
	if(iI > iJ){
	  Swap(&iI,&iJ);
	}
	if(iI > iK){
	  Swap(&iI,&iK);
	}
	if(iJ > iK){
	  Swap(&iJ,&iK);
	}
	//printf("Invtransposition1, good i: %d, j: %d, k: %d\n",iI,iJ,iK);
	Invtransposition1(iPermutation,iI,iJ,iK);
	(*iN)++;
      }
      else if(iTemp < iGoodtrans + iGoodinvtrans1 + iGoodinvtrans2){
	iTemp = iIntrand(iGoodinvtrans2);
	iI = iGoodinvtranslist2[iTemp][0];
	iJ = iGoodinvtranslist2[iTemp][1];
	iK = iGoodinvtranslist2[iTemp][2];
	if(iI > iJ){
	  Swap(&iI,&iJ);
	}
	if(iI > iK){
	  Swap(&iI,&iK);
	}
	if(iJ > iK){
	  Swap(&iJ,&iK);
	}
	//printf("Invtransposition2, good i: %d, j: %d, k: %d\n",iI,iJ,iK);
	Invtransposition2(iPermutation,iI,iJ,iK);
	(*iN)++;
      }
      else{
	/* +1-inversion */
	iTemp = iIntrand(iKumPlus1[iNumberofcycles - 1]);
	//	printf("Temp:%d\n",iTemp);
	//	for(iI = 0; iI < iNumberofcycles; iI++){
	// printf("iKumPlus1[i]: %d\n",iKumPlus1[iI]);
	//}
	for(iK = 0; iTemp >= iKumPlus1[iK]; iK++);
	//printf("k: %d\n",iK);
	//printf("iCyclelength[iK]: %d\n",iCyclelength[iK]);
	//printf("iNumberofplus[iK]: %d\n",iNumberofplus[iK]);

	iI = iListofnegatives[iK][iIntrand(iCyclelength[iK] - iNumberofplus[iK])];
	iJ = iListofpositives[iK][iIntrand(iNumberofplus[iK])];
	if(iI > iJ){
	  Swap(&iI,&iJ);
	}
	//printf("Inversion, +1 i: %d, j: %d\n",iI,iJ);
	Reversion(iPermutation,iI,iJ);
	(*iM)++;
       }
    }
  }
  /* 0-inversion */
  else if((iReversesame > 0 && iIntrand(100) < 50) || (iReverseminus == 0 && 2 * (iGoodall + iSuperGoodall) == XX * XX * XX - XX)){
    if(iGoodall + iSuperGoodall + iReverseplus != 0){
      eProposal *= 0.01;
    }
    if(iReverseminus != 0 || 2*(iGoodall + iSuperGoodall) != XX * XX * XX - XX){
      eProposal *= 0.5;
    }
    iTemp = iIntrand(iKum0[iNumberofcycles - 1]);
    for(iK = 0; iTemp >= iKum0[iK]; iK++);
    iTemp = iIntrand(iNumberofplus[iK] * (iNumberofplus[iK] - 1) + (iCyclelength[iK] - iNumberofplus[iK]) * (iCyclelength[iK] - iNumberofplus[iK] - 1) );
    if(iTemp < iNumberofplus[iK] * (iNumberofplus[iK] - 1)){
      iI = iListofpositives[iK][iIntrand(iNumberofplus[iK])];
      iJ = iI;
      while(iJ == iI){
	iJ = iListofpositives[iK][iIntrand(iNumberofplus[iK])];
      }
      if(iI > iJ){
	Swap(&iI,&iJ);
      }
      //printf("Inversion, 0 i: %d, j: %d\n",iI,iJ);
      Reversion(iPermutation,iI,iJ);
      (*iM)++;
    }
    else{
      iI = iListofnegatives[iK][iIntrand(iCyclelength[iK] - iNumberofplus[iK])];
      iJ = iI;
      while(iJ == iI){
	iJ = iListofnegatives[iK][iIntrand(iCyclelength[iK] - iNumberofplus[iK])];
      }
      if(iI > iJ){
	Swap(&iI,&iJ);
      }
      //printf("Inversion, 0' i: %d, j: %d\n",iI,iJ);
      Reversion(iPermutation,iI,iJ);
      (*iM)++;
    }
  }
  /* -1-inversion */
  else if(iReverseminus > 0 && iIntrand(100) < 80){
    eProposal = 0.8;
    if(iGoodall + iSuperGoodall + iReverseplus != 0){
      eProposal *= 0.01;
    }
    if(iReversesame > 0){
      eProposal *= 0.5;
    }
    eProposal /= (double) iReverseminus;
    /*
    iTemp = iIntrand(iKumMinus11[iNumberofcycles - 1]);
    for(iK = 0; iTemp >= iKumMinus11[iK]; iK++);
    iTemp = iIntrand(iCyclelength[iK]);
    if(iTemp < iNumberofplus[iK]){
      iI = iListofpositives[iK][iIntrand(iNumberofplus[iK])];
    }
    else{
      iI = iListofnegatives[iK][iIntrand(iCyclelength[iK] - iNumberofplus[iK])];
    }
    iTemp = iIntrand(iKumMinus12[iNumberofcycles - 1] - iKumMinus12[iK]);
    for(iL = 0; iTemp >= iKumMinus12[iL]; iL++){
      if(iL == iK - 1){
	iTemp += iCyclelength[iK];
      }
    }
    iTemp = iIntrand(iCyclelength[iL]);
    if(iTemp < iNumberofplus[iL]){
      iJ = iListofpositives[iL][iIntrand(iNumberofplus[iL])];
    }
    else{
      iJ = iListofnegatives[iL][iIntrand(iCyclelength[iL] - iNumberofplus[iL])];
    }
    if(iI > iJ){
      Swap(&iI,&iJ);
    }
    */
    //printf("Inversion, -1 i: %d, j: %d\n",iI,iJ);
    iAccepted = 0;
    while(iAccepted == 0){
      iI = iIntrand(XX + 1);
      iJ = iIntrand(XX);
      if(iJ == iI){
	iJ++;
      }
      if(iI > iJ){
	Swap(&iI,&iJ);
      }
      if(iCyclenumber[iI] != iCyclenumber[iJ]){
	iAccepted = 1;
      }
    }
    Reversion(iPermutation,iI,iJ);
    (*iM)++;
  }
  /* non-benign transposition */
  else{
    if(iGoodall + iSuperGoodall + iReverseplus != 0){
      eProposal *= 0.01;
    }
    if(iReversesame > 0){
      eProposal *= 0.5;
    }
    if(iReverseminus > 0){
      eProposal *= 0.2;
    }

    /**********************************************************************************************/
      iBadtrans=((XX+1)*XX*(XX-1)/6)-iGoodtrans - iSuperGoodtrans;
      iBadinvtrans1=((XX+1)*XX*(XX-1)/6)-iGoodinvtrans1 - iSuperGoodinvtrans1;
      iBadinvtrans2=((XX+1)*XX*(XX-1)/6)-iGoodinvtrans2 - iSuperGoodinvtrans2;

      iAccepted=0;
      if(iIntrand(iBadtrans+iBadinvtrans1+iBadinvtrans2)<iBadtrans){
	while(iAccepted==0){
	  /* transposition */
	  iI=iIntrand(XX+1);
	  iJ=iIntrand(XX);
	  if(iJ>=iI){
	    iJ++;
	  }
	  else{
	    Swap(&iI,&iJ);
	  }
	  iK=iIntrand(XX-1);
	  if(iK>=iI)
	    iK++;
	  if(iK>=iJ)
	    iK++;
	  if(iK<iI){
	    Swap(&iK,&iI);
	    Swap(&iK,&iJ);
	  }
	  if(iK<iJ){
	    Swap(&iK,&iJ);
	  }	  
	  iAccepted=1;
	  for(iTemp=0;iTemp<iGoodtrans;iTemp++){
	    if(iI==iGoodtranslist[iTemp][0]&&iJ==iGoodtranslist[iTemp][1]&&iK==iGoodtranslist[iTemp][2])
	      iAccepted=0;
	    if(iI==iGoodtranslist[iTemp][0]&&iK==iGoodtranslist[iTemp][1]&&iJ==iGoodtranslist[iTemp][2])
	      iAccepted=0;
	    if(iJ==iGoodtranslist[iTemp][0]&&iI==iGoodtranslist[iTemp][1]&&iK==iGoodtranslist[iTemp][2])
	      iAccepted=0;
	    if(iJ==iGoodtranslist[iTemp][0]&&iK==iGoodtranslist[iTemp][1]&&iI==iGoodtranslist[iTemp][2])
	      iAccepted=0;
	    if(iK==iGoodtranslist[iTemp][0]&&iI==iGoodtranslist[iTemp][1]&&iJ==iGoodtranslist[iTemp][2])
	      iAccepted=0;
	    if(iK==iGoodtranslist[iTemp][0]&&iJ==iGoodtranslist[iTemp][1]&&iI==iGoodtranslist[iTemp][2])
	      iAccepted=0;
	  }
	  for(iTemp=0;iTemp<iSuperGoodtrans;iTemp++){
	    if(iI==iSuperGoodtranslist[iTemp][0]&&iJ==iSuperGoodtranslist[iTemp][1]&&iK==iSuperGoodtranslist[iTemp][2])
	      iAccepted=0;
	    if(iI==iSuperGoodtranslist[iTemp][0]&&iK==iSuperGoodtranslist[iTemp][1]&&iJ==iSuperGoodtranslist[iTemp][2])
	      iAccepted=0;
	    if(iJ==iSuperGoodtranslist[iTemp][0]&&iI==iSuperGoodtranslist[iTemp][1]&&iK==iSuperGoodtranslist[iTemp][2])
	      iAccepted=0;
	    if(iJ==iSuperGoodtranslist[iTemp][0]&&iK==iSuperGoodtranslist[iTemp][1]&&iI==iSuperGoodtranslist[iTemp][2])
	      iAccepted=0;
	    if(iK==iSuperGoodtranslist[iTemp][0]&&iI==iSuperGoodtranslist[iTemp][1]&&iJ==iSuperGoodtranslist[iTemp][2])
	      iAccepted=0;
	    if(iK==iSuperGoodtranslist[iTemp][0]&&iJ==iSuperGoodtranslist[iTemp][1]&&iI==iSuperGoodtranslist[iTemp][2])
	      iAccepted=0;
	  }
	  if(iAccepted==1){
	    //printf("Transposition, bad, i: %d, j: %d, k: %d\n",iI,iJ,iK);
	    Transposition(iPermutation,iI,iJ,iK);
	    /* printf("transposition, %d, %d, %d\n",i,iJ,iK);*/
	    (*iN)++;
	    eProposal *= 1.0/(double)(iBadtrans+iBadinvtrans1+iBadinvtrans2);
	  }
	}
      }
      else if(iIntrand(iBadinvtrans1+iBadinvtrans2)<iBadinvtrans1){
	while(iAccepted==0){
	  /* invtransposition1 */
	  iI=iIntrand(XX+1);
	  iJ=iIntrand(XX);
	  if(iJ>=iI){
	    iJ++;
	  }
	  else{
	    Swap(&iI,&iJ);
	  }
	  iK=iIntrand(XX-1);
	  if(iK>=iI){
	    iK++;
	  }
	  if(iK>=iJ){
	    iK++;
	  }
	  if(iK<iI){
	    Swap(&iI,&iK);
	    Swap(&iJ,&iK);
	  }
	  if(iK<iJ){
	    Swap(&iJ,&iK);
	  }	  
	  iAccepted=1;
	  for(iTemp=0;iTemp<iGoodinvtrans1;iTemp++){
	    if(iI==iGoodinvtranslist1[iTemp][0]&&iJ==iGoodinvtranslist1[iTemp][1]&&iK==iGoodinvtranslist1[iTemp][2])
	      iAccepted=0;
	    if(iI==iGoodinvtranslist1[iTemp][0]&&iK==iGoodinvtranslist1[iTemp][1]&&iJ==iGoodinvtranslist1[iTemp][2])
	      iAccepted=0;
	    if(iJ==iGoodinvtranslist1[iTemp][0]&&iI==iGoodinvtranslist1[iTemp][1]&&iK==iGoodinvtranslist1[iTemp][2])
	      iAccepted=0;
	    if(iJ==iGoodinvtranslist1[iTemp][0]&&iK==iGoodinvtranslist1[iTemp][1]&&iI==iGoodinvtranslist1[iTemp][2])
	      iAccepted=0;
	    if(iK==iGoodinvtranslist1[iTemp][0]&&iI==iGoodinvtranslist1[iTemp][1]&&iJ==iGoodinvtranslist1[iTemp][2])
	      iAccepted=0;
	    if(iK==iGoodinvtranslist1[iTemp][0]&&iJ==iGoodinvtranslist1[iTemp][1]&&iI==iGoodinvtranslist1[iTemp][2])
	      iAccepted=0;
	  }
	  for(iTemp=0;iTemp<iSuperGoodinvtrans1;iTemp++){
	    if(iI==iSuperGoodinvtranslist1[iTemp][0]&&iJ==iSuperGoodinvtranslist1[iTemp][1]&&iK==iSuperGoodinvtranslist1[iTemp][2])
	      iAccepted=0;
	    if(iI==iSuperGoodinvtranslist1[iTemp][0]&&iK==iSuperGoodinvtranslist1[iTemp][1]&&iJ==iSuperGoodinvtranslist1[iTemp][2])
	      iAccepted=0;
	    if(iJ==iSuperGoodinvtranslist1[iTemp][0]&&iI==iSuperGoodinvtranslist1[iTemp][1]&&iK==iSuperGoodinvtranslist1[iTemp][2])
	      iAccepted=0;
	    if(iJ==iSuperGoodinvtranslist1[iTemp][0]&&iK==iSuperGoodinvtranslist1[iTemp][1]&&iI==iSuperGoodinvtranslist1[iTemp][2])
	      iAccepted=0;
	    if(iK==iSuperGoodinvtranslist1[iTemp][0]&&iI==iSuperGoodinvtranslist1[iTemp][1]&&iJ==iSuperGoodinvtranslist1[iTemp][2])
	      iAccepted=0;
	    if(iK==iSuperGoodinvtranslist1[iTemp][0]&&iJ==iSuperGoodinvtranslist1[iTemp][1]&&iI==iSuperGoodinvtranslist1[iTemp][2])
	      iAccepted=0;
	  }
	  if(iAccepted==1){
	    //printf("Invtransposition1, bad, i: %d, j: %d, k: %d\n",iI,iJ,iK);
	    Invtransposition1(iPermutation,iI,iJ,iK);
	    /* printf("invtransposition1, %d, %d, %d\n",i,iJ,iK);*/
	    (*iN)++;
	    eProposal *= 1.0/(double)(iBadtrans+iBadinvtrans1+iBadinvtrans2);
	  }
	}
      }
      else{
	iAccepted=1;
	iI=iIntrand(XX+1);
	iJ=iIntrand(XX);
	if(iJ>=iI){
	  iJ++;
	}
	else{
	  Swap(&iI,&iJ);
	}
	iK=iIntrand(XX-1);
	if(iK>=iI){
	  iK++;
	}
	if(iK>=iJ){
	  iK++;
	}
	if(iK<iI){
	  Swap(&iI,&iK);
	  Swap(&iJ,&iK);
	}
	if(iK<iJ){
	  Swap(&iJ,&iK);
	}	  	
	for(iTemp=0;iTemp<iGoodinvtrans2;iTemp++){
	  if(iI==iGoodinvtranslist2[iTemp][0]&&iJ==iGoodinvtranslist2[iTemp][1]&&iK==iGoodinvtranslist2[iTemp][2])
	    iAccepted=0;
	  if(iI==iGoodinvtranslist2[iTemp][0]&&iK==iGoodinvtranslist2[iTemp][1]&&iJ==iGoodinvtranslist2[iTemp][2])
	    iAccepted=0;
	  if(iJ==iGoodinvtranslist2[iTemp][0]&&iI==iGoodinvtranslist2[iTemp][1]&&iK==iGoodinvtranslist2[iTemp][2])
	    iAccepted=0;
	  if(iJ==iGoodinvtranslist2[iTemp][0]&&iK==iGoodinvtranslist2[iTemp][1]&&iI==iGoodinvtranslist2[iTemp][2])
	    iAccepted=0;
	  if(iK==iGoodinvtranslist2[iTemp][0]&&iI==iGoodinvtranslist2[iTemp][1]&&iJ==iGoodinvtranslist2[iTemp][2])
	    iAccepted=0;
	  if(iK==iGoodinvtranslist2[iTemp][0]&&iJ==iGoodinvtranslist2[iTemp][1]&&iI==iGoodinvtranslist2[iTemp][2])
	    iAccepted=0;
	}
	for(iTemp=0;iTemp<iSuperGoodinvtrans2;iTemp++){
	  if(iI==iSuperGoodinvtranslist2[iTemp][0]&&iJ==iSuperGoodinvtranslist2[iTemp][1]&&iK==iSuperGoodinvtranslist2[iTemp][2])
	    iAccepted=0;
	  if(iI==iSuperGoodinvtranslist2[iTemp][0]&&iK==iSuperGoodinvtranslist2[iTemp][1]&&iJ==iSuperGoodinvtranslist2[iTemp][2])
	    iAccepted=0;
	  if(iJ==iSuperGoodinvtranslist2[iTemp][0]&&iI==iSuperGoodinvtranslist2[iTemp][1]&&iK==iSuperGoodinvtranslist2[iTemp][2])
	    iAccepted=0;
	  if(iJ==iSuperGoodinvtranslist2[iTemp][0]&&iK==iSuperGoodinvtranslist2[iTemp][1]&&iI==iSuperGoodinvtranslist2[iTemp][2])
	    iAccepted=0;
	  if(iK==iSuperGoodinvtranslist2[iTemp][0]&&iI==iSuperGoodinvtranslist2[iTemp][1]&&iJ==iSuperGoodinvtranslist2[iTemp][2])
	    iAccepted=0;
	  if(iK==iSuperGoodinvtranslist2[iTemp][0]&&iJ==iSuperGoodinvtranslist2[iTemp][1]&&iI==iSuperGoodinvtranslist2[iTemp][2])
	    iAccepted=0;
	}
	if(iAccepted==1){
	  //printf("Invtransposition2, bad, i: %d, j: %d, k: %d\n",iI,iJ,iK);
	  Invtransposition2(iPermutation,iI,iJ,iK);
	  /*		  printf("invtransposition2, %d, %d, %d\n",i,iJ,iK);*/
	  (*iN)++;
	  eProposal *= 1.0/(double)(iBadtrans+iBadinvtrans1+iBadinvtrans2);
	}
      }
      
      /**********************************************************************************************/
  }

  return(eProposal);
}

void Reversion(int iPermutation[2 * XX + 2],
	       int iI,
	       int iJ){

  int iK,iTempperm[2 * XX + 2];

  for(iK = 2 * iI + 1; iK < 2 * iJ + 1; iK++){
    iTempperm[iK] = iPermutation[2 * iJ + 2 * iI + 1 - iK];
  }
  for(iK = 2 * iI + 1; iK < 2 * iJ + 1; iK++){
    iPermutation[iK]=iTempperm[iK];
  }
}

void Transposition(int iPermutation[2 * XX + 2],
		   int iI,
		   int iJ,
		   int iK){

  int iTemp, iTempperm[2 * XX + 2];

  for(iTemp = 2 * iI + 1; iTemp < 2 * iJ + 1; iTemp++){
    iTempperm[2 * iK + iTemp - 2 * iJ] = iPermutation[iTemp];
  }
  for(iTemp = 2 * iJ + 1; iTemp < 2 * iK + 1; iTemp++){
    iTempperm[2 * iI + iTemp - 2 * iJ] = iPermutation[iTemp];
  }
  for(iTemp = 2 * iI + 1; iTemp < 2 * iK + 1; iTemp++){
    iPermutation[iTemp] = iTempperm[iTemp];
  }

}


void Invtransposition1(int iPermutation[2 * XX + 2],
		       int iI,
		       int iJ,
		       int iK){

  int iTemp, iTempperm[2 * XX + 2];

  for(iTemp = 2 * iI + 1; iTemp < 2 * iJ + 1; iTemp++){
    iTempperm[2 * iK + iTemp - 2 * iJ] = iPermutation[2 * iJ + 2 * iI + 1 - iTemp];
  }
  for(iTemp = 2 * iJ + 1; iTemp < 2 * iK + 1; iTemp++){
    iTempperm[2 * iI + iTemp - 2 * iJ] = iPermutation[iTemp];
  }
  for(iTemp = 2 * iI + 1; iTemp < 2 * iK + 1; iTemp++){
    iPermutation[iTemp] = iTempperm[iTemp];
  }

}

void Invtransposition2(int iPermutation[2 * XX + 2],
		       int iI,
		       int iJ,
		       int iK){

  int iTemp, iTempperm[2 * XX + 2];
	
  for(iTemp = 2 * iI + 1; iTemp < 2 * iJ + 1; iTemp++){
    iTempperm[2 * iK + iTemp - 2 * iJ] = iPermutation[iTemp];
  }
  for(iTemp = 2 * iJ + 1; iTemp < 2 * iK + 1; iTemp++){
    iTempperm[2 * iI + iTemp - 2 * iJ] = iPermutation[2 * iK + 2 * iJ + 1 - iTemp];
  }
  for(iTemp = 2 * iI + 1; iTemp < 2 * iK + 1; iTemp++){
    iPermutation[iTemp] = iTempperm[iTemp];
  }
}



int iIntrand(int iN){

  double eZ;

//++pac
#ifdef _PLAN9
  eZ = (double)rand() * (double)iN / (double)RAND_MAX;
#else
  eZ = (double)random() * (double)iN / (double)RAND_MAX;
#endif

  if(eZ >= iN)
    eZ = iN - 1;
  if(eZ < 0)
    eZ = 0;
  return((int)floor(eZ));
}

void Swap(int *iI, int *iJ){

  int iK;

  iK = *iI;
  *iI = *iJ;
  *iJ = iK;
}
